
; x86-state.lisp                               Warren A. Hunt, Jr.

; We model the X86 state with two arrays: one array of 32-bit values
; for the registers and array of 8-bit values for the memory.  There
; are a large number of registers, as this data structure holds all of
; the supervisor data as well.

(in-package "ACL2")

(include-book "constants")

; Increase memory for X86 memory.
(include-book "centaur/misc/memory-mgmt-logic" :dir :system)
(value-triple (set-max-mem (* 6 (expt 2 30))))

; Here we include the GL book to help verify various arithmetic facts.
(local (include-book "centaur/gl/gl" :dir :system))

(defun gl-int (start by count)
  (declare (xargs :guard (and (natp start)
                              (natp by)
                              (natp count))))
  (if (zp count)
      nil
    (cons start
          (gl-int (+ by start) by (1- count)))))


(defstobj x86-32

; Notice that the array fields use UNSIGNED-BYTE types rather than
; types (SATISFIES N32P) and (SATISFIES N08P).  The reason is that we
; want to help the host Lisp system allocate these arrays using
; minimal real (physical) memory.

  ;; The model register file has a simple structure; it just an array
  ;; of 32-bit words.
  (rgf :type (array (unsigned-byte 32)
                    (*m86-32-reg-names-len*))
       :initially 0
       :resizable nil)
  ;; The program counter.
  (eip :type (unsigned-byte 32)
       :initially 0)
  ;; The processor flag register -- modeled after the X86.
  (flg :type (unsigned-byte 32)
       :initially 0)
  ;; The segment registers.
  (seg :type (array (unsigned-byte 32)
                    (*m86-32-segment-reg-names-len*))
       :initially 0
       :resizable nil)
  ;; The memory-management table registers.
  (tbl :type (array (unsigned-byte 16)
                    (*m86-32-table-reg-names-len*))
       :initially 0
       :resizable nil)
  ;; The control registers
  (cr  :type (array (unsigned-byte 32)
                    (*m86-32-control-reg-names-len*))
       :initially 0
       :resizable nil)

  ;; Additional registers to be defined here.

  ;; FPU 16-bit status registers

  ;; FPU 48-bit last instruction registers

  ;; FPU 80-bit data registers

  ;; XMM 16-bit status

  ;; XMM 128-bit data registers

  ;; The model memory is intended to represent a 2^32 byte physical
  ;; memory.  Starting with the PAE extension, X86 processors can
  ;; address more than 2^32 bytes, but we aren't modeling that as yet.
  ;; We choose (for performance reasons) to model the 2^32 byte memory
  ;; as 2^30 double-word (4-byte) memory.  We maintain the byte-level
  ;; semantics.
  (mem :type (array (unsigned-byte 32)
                    (*mem-size-in-dwords*))
       :initially 0
       :resizable nil)
  ;; The state of the ACL2 model.  This flag is not part of the X86
  ;; processor; it is used to signal problems with model state, such
  ;; as the processor is halted.  While this flag is NIL, the processor
  ;; model is OK; otherwise, the flag indicates the problem.
  (ms :type t :initially nil)
  :inline t
  )

; Event to observe all events introduced by DEFSTOBJ.

(defun get-stobj-raw-defs (form state)
  (declare (xargs :mode :program :stobjs (state)))
  (let* ((name (cadr form))
         (args (cddr form))
         (wrld (w state))
         (template (defstobj-template name args wrld)))
    (defstobj-raw-defs name template wrld)))

; Lemmas to help with proofs about STOBJs that hold X86 state.  Our
; goal is to prove a nice set of forward-chaining lemmas, as our
; predicates seem nicely set up for that.

(in-theory (disable nth))  ; Because NTH used to select object from
                           ; the x86-32 state.

; Lemmas below are in the same order as the fields declare in the
; X86-32 STOBJ above.

; We first deal with the STOBJ read lemmas

; RGF read lemmas

(defthm natp-nth-of-rgf
  (implies (and (rgfp x)
                (integerp i)
                (<= 0 i)
                (< i (len x)))
           (and (integerp (nth i x))
                (<= 0 (nth i x))))
  :rule-classes :type-prescription
  :hints (("Goal" :in-theory (e/d (nth) ()))))

(defthm nth-of-rgf-<=4294967296
  (implies (and (rgfp x)
                (integerp i)
                (<= 0 i)
                (< i (len x)))
           (< (nth i x) 4294967296))
  :rule-classes :linear
  :hints (("Goal" :in-theory (e/d (nth) ()))))

(defthm natp-rgfi
  (implies (and (force (x86-32p x86-32))
                (force (n03p i)))
           (and (integerp (rgfi i x86-32))
                (<= 0 (rgfi i x86-32))))
  :rule-classes :type-prescription)

(defthm rgfi-less-than-expt-2-32
  (implies (and (x86-32p x86-32)
                (n03p i))
           (< (rgfi i x86-32) 4294967296))
  :rule-classes :linear)


; EIP read lemmas

(defthm natp-eip
  (implies (force (x86-32p x86-32))
           (and (integerp (eip x86-32))
                (<= 0 (eip x86-32))))
  :rule-classes :type-prescription)

(defthm eip-less-than-expt-2-32
  (implies (x86-32p x86-32)
           (< (eip x86-32) 4294967296))
  :rule-classes :linear)


; FLG read lemmas

(defthm natp-flg
  (implies (force (x86-32p x86-32))
           (and (integerp (flg x86-32))
                (<= 0 (flg x86-32))))
  :rule-classes :type-prescription)

(defthm flg-less-than-expt-2-32
  (implies (x86-32p x86-32)
           (< (flg x86-32) 4294967296))
  :rule-classes :linear)


; SEG read lemmas

(defthm natp-nth-of-seg
  (implies (and (segp x)
                (integerp i)
                (<= 0 i)
                (< i (len x)))
           (and (integerp (nth i x))
                (<= 0 (nth i x))))
  :rule-classes :type-prescription
  :hints (("Goal" :in-theory (e/d (nth) ()))))

(defthm nth-of-seg-<=4294967296
  (implies (and (segp x)
                (integerp i)
                (<= 0 i)
                (< i (len x)))
           (< (nth i x) 4294967296))
  :rule-classes :linear
  :hints (("Goal" :in-theory (e/d (nth) ()))))

(defthm natp-segi
  (implies (and (force (x86-32p x86-32))
                (force (n03p i))
                (force (< i 6)))
           (and (integerp (segi i x86-32))
                (<= 0 (segi i x86-32))))
  :rule-classes :type-prescription)

(defthm segi-less-than-expt-2-32
  (implies (and (x86-32p x86-32)
                (n03p i)
                (< i 6))
           (< (segi i x86-32) 4294967296))
  :rule-classes :linear)


; TBL read lemmas

(defthm natp-nth-of-tbl
  (implies (and (tblp x)
                (integerp i)
                (<= 0 i)
                (< i (len x)))
           (and (integerp (nth i x))
                (<= 0 (nth i x))))
  :rule-classes :type-prescription
  :hints (("Goal" :in-theory (e/d (nth) ()))))

(defthm nth-of-tbl-<=65536
  (implies (and (tblp x)
                (integerp i)
                (<= 0 i)
                (< i (len x)))
           (< (nth i x) 65536))
  :rule-classes :linear
  :hints (("Goal" :in-theory (e/d (nth) ()))))

(defthm natp-tbli
  (implies (and (force (x86-32p x86-32))
                (force (n02p i)))
           (and (integerp (tbli i x86-32))
                (<= 0 (tbli i x86-32))))
  :rule-classes :type-prescription)

(defthm tbli-less-than-expt-2-32
  (implies (and (x86-32p x86-32)
                (n02p i)
                (< i 6))
           (< (tbli i x86-32) 65536))
  :rule-classes :linear)


; CR read lemmas

(defthm natp-nth-of-cr
  (implies (and (crp x)
                (integerp i)
                (<= 0 i)
                (< i (len x)))
           (and (integerp (nth i x))
                (<= 0 (nth i x))))
  :rule-classes :type-prescription
  :hints (("Goal" :in-theory (e/d (nth) ()))))

(defthm nth-of-cr-<=4294967296
  (implies (and (crp x)
                (integerp i)
                (<= 0 i)
                (< i (len x)))
           (< (nth i x) 4294967296))
  :rule-classes :linear
  :hints (("Goal" :in-theory (e/d (nth) ()))))

(defthm natp-cri
  (implies (and (force (x86-32p x86-32))
                (force (n03p i))
                (force (< i 6)))
           (and (integerp (cri i x86-32))
                (<= 0 (cri i x86-32))))
  :rule-classes :type-prescription)

(defthm cri-less-than-expt-2-32
  (implies (and (x86-32p x86-32)
                (n03p i)
                (< i 6))
           (< (cri i x86-32) 4294967296))
  :rule-classes :linear)


; MEM read lemmas

(defthm natp-nth-of-mem
  (implies (and (memp x)
                (integerp i)
                (<= 0 i)
                (< i (len x)))
           (and (integerp (nth i x))
                (<= 0 (nth i x))))
  :rule-classes :type-prescription
  :hints (("Goal" :in-theory (e/d (nth)))))

(defthm nth-of-mem-<=-4294967296
  (implies (and (memp x)
                (integerp i)
                (<= 0 i)
                (< i (len x)))
           (< (nth i x) 4294967296))
  :rule-classes :linear
  :hints (("Goal" :in-theory (e/d (nth) ()))))

(defthm natp-memi-when-n30p-addr
  (implies (and (force (x86-32p x86-32))
                (force (n30p addr)))
           (and (integerp (memi addr x86-32))
                (<= 0 (memi addr x86-32))))
  :rule-classes (:rewrite :type-prescription))

(defthm memi-less-than-expt-2-32
  (implies (and (x86-32p x86-32)
                (n30p addr))
           (< (memi addr x86-32) 4294967296))
  :rule-classes :linear)


(encapsulate
 ()

 (local
  (def-gl-thm n32p-ash--2-is-n30p
    :hyp (and (integerp x)
              (<= 0 x)
              (< x 4294967296))
    :concl (and (<= 0 (ash x -2))
                (< (ash x -2) 1073741824))
    :g-bindings
    `((x (:g-number ,(gl-int  0  1  33))))))

 (defthm natp-memi-when-n32p-addr
   (implies (and (force (x86-32p x86-32))
                 (force (n32p addr)))
            (and (integerp (memi (ash addr -2) x86-32))
                 (<= 0 (memi (ash addr -2) x86-32))))
   :rule-classes :type-prescription
   :hints (("Goal" :in-theory (e/d () (memi x86-32p)))))

 (defthm memi-less-than-expt-2-32-when-n32p-addr
   (implies (and (x86-32p x86-32)
                 (n32p addr))
            (< (memi (ash addr -2) x86-32) 4294967296))
   :rule-classes :linear
   :hints (("Goal" :in-theory (e/d () (memi x86-32p))))))


; We wonder if the two lemmas about UPDATE-xxx would be better as
; :FORWARD-CHAINING rules (or, as both :REWRITE and :FORWARD-CHAINING
; rules), using *MEM-SIZE-IN-BYTES* and *REG-SIZE-IN-DWRDS* in the
; hypotheses instead of LEN.

(defthm length-is-len-when-not-stringp
  (implies (not (stringp x))
           (equal (length x)
                  (len x)))
  :hints (("Goal" :in-theory (e/d (length) ()))))


; RGF update lemmas

(defthm rgfp-update-nth
  (implies (and (rgfp x)
                (integerp i)
                (<= 0 i)
                (< i (len x))
                (n32p v))
           (rgfp (update-nth i v x))))

(defthm x86-32p-update-rgfi-n03p
  (implies (and (x86-32p x86-32)
                (n03p i)
                (n32p v))
           (x86-32p (update-rgfi i v x86-32))))


; EIP update lemma

(defthm x86-32p-update-eip
  (implies (and (x86-32p x86-32)
                (n32p v))
           (x86-32p (update-eip v x86-32))))


; EFLAGS update lemma

(defthm x86-32p-update-flg
  (implies (and (x86-32p x86-32)
                (n32p v))
           (x86-32p (update-flg v x86-32))))


; SEG register update lemmas

(defthm segp-update-nth
  (implies (and (segp x)
                (integerp i)
                (<= 0 i)
                (< i (len x))
                (n32p v))
           (segp (update-nth i v x))))

(defthm x86-32p-update-segi
  (implies (and (x86-32p x86-32)
                (integerp i)
                (<= 0 i)
                (< i 6)
                (n32p v))
           (x86-32p (update-segi i v x86-32))))


; TBL update lemmas

(defthm tblp-update-nth
  (implies (and (tblp x)
                (integerp i)
                (<= 0 i)
                (< i (len x))
                (n16p v))
           (tblp (update-nth i v x))))

(defthm x86-32p-update-tbli
  (implies (and (x86-32p x86-32)
                (n02p i)
                (n16p v))
           (x86-32p (update-tbli i v x86-32))))


; CR register update lemmas

(defthm crp-update-nth
  (implies (and (crp x)
                (integerp i)
                (<= 0 i)
                (< i (len x))
                (n32p v))
           (crp (update-nth i v x))))

(defthm x86-32p-update-cri
  (implies (and (x86-32p x86-32)
                (integerp i)
                (<= 0 i)
                (< i 6)
                (n32p v))
           (x86-32p (update-cri i v x86-32))))


; Memory update lemmas

(defthm memp-update-nth
  (implies (and (memp x)
                (integerp i)
                (<= 0 i)
                (< i (len x))
                (n32p v))
           (memp (update-nth i v x))))

(defthm x86-32p-update-memi-n30p
  (implies (and (force (x86-32p x86-32))
                (force (n30p i))
                (force (n32p v)))
           (x86-32p (update-memi i v x86-32))))

(encapsulate
 ()

 (local
  (def-gl-thm n32p-ash--2-is-n30p
    :hyp (n32p x)
    :concl (n30p (ash x -2))
    :g-bindings
     `((x (:g-number ,(gl-int  0  1  33))))))

 (defthm x86-32p-update-memi-when-addr-is-n32p
   (implies (and (x86-32p x86-32)
                 (n32p addr)
                 (n32p v))
            (x86-32p (update-memi (ash addr -2) v x86-32)))
   :hints (("Goal" :in-theory (e/d () (update-memi))))))


; Should we have this next group of lemmas?  Probably not, but the
; first, for instance, one allows later lemma X86-32P-WMB-NO-WRAP (see
; below) to be proven.

(defthm len-x86-32-rgf
  (implies (x86-32p x86-32)
           (equal (len (nth *rgfi* x86-32))
                  *m86-32-reg-names-len*)))

(defthm len-x86-32-seg
  (implies (x86-32p x86-32)
           (equal (len (nth *segi* x86-32))
                  *m86-32-segment-reg-names-len*)))

(defthm len-x86-32-tbl
  (implies (x86-32p x86-32)
           (equal (len (nth *tbli* x86-32))
                  *m86-32-table-reg-names-len*)))

(defthm len-x86-32-mem
  (implies (x86-32p x86-32)
           (equal (len (nth *memi* x86-32))
                  *mem-size-in-dwords*)))

(defthm x86-32p-properties
  (implies (x86-32p x86-32)
           (and (true-listp x86-32)
                (equal (len x86-32) 8)

                (rgfp (nth 0 x86-32))
                (equal (len (nth 0 x86-32))
                       *m86-32-reg-names-len*)

                (eipp (nth 1 x86-32))
                (flgp (nth 2 x86-32))

                (segp (nth 3 x86-32))
                (equal (len (nth 3 x86-32))
                       *m86-32-segment-reg-names-len*)

                (tblp (nth 4 x86-32))
                (equal (len (nth 4 x86-32))
                       *m86-32-table-reg-names-len*)

                (crp (nth 5 x86-32))
                (equal (len (nth 5 x86-32))
                       *m86-32-control-reg-names-len*)

                (memp (nth 6 x86-32))
                (equal (len (nth 6 x86-32))
                       *mem-size-in-dwords*)))
  :hints (("Goal" :in-theory (enable x86-32p rgfp eipp flgp segp tblp crp memp)))
  :rule-classes :forward-chaining)

; Additional lemmas to help with later guard proofs.

; Hopefully, we have proven all the facts we need about the X86
; machine state.

(in-theory (disable x86-32p
                    rgfp rgfi update-rgfi
                    eipp eip  update-eip
                    flgp flg  update-flg
                    segp segi update-segi
                    tblp tbli update-tbli
                    memp memi update-memi))


; Translate virtual address into physical address.

(defun vaddr-to-paddr (addr x86-32)
  (declare (xargs :guard (n32p addr)
                  :stobjs (x86-32)))
  (declare (ignore x86-32))
  addr)


; Read memory byte and memory double-word functions

(encapsulate
 ()
 (local
  (def-gl-thm n32p-ash--2-is-n30p
    :hyp (n32p x)
    :concl (n30p (ash x -2))
    :g-bindings
     `((x (:g-number ,(gl-int  0 1 33))))))

 (local
  (def-gl-thm n32p-ash-x-is-less-than-n30p
    :hyp (n32p addr)
    :concl (< (ash addr (- 2)) 1073741824)
    :rule-classes :linear
    :g-bindings
     `((addr (:g-number ,(gl-int  0 1 33))))))

 (local
  (def-gl-thm ash-logand-addr-3-is-integerp-less-or-equal-24
    :hyp (n32p addr)
    :concl (<= (ash (logand addr 3) 3) 24)
    :g-bindings
     `((addr (:g-number ,(gl-int  0 1 33))))))

 (local
  (def-gl-thm ash-memi-ash-logand-addr
    :hyp (and (n32p mem-value)
              (n32p addr))
    :concl (< (ash mem-value (- (ash (logand addr 3) 3)))
              4294967296)
    :rule-classes :linear
    :g-bindings
    `((addr      (:g-number ,(gl-int  0 1 33)))
      (mem-value (:g-number ,(gl-int 33 1 33))))))

 (defun rm08 (addr x86-32)
   (declare (xargs :guard (n32p addr)
                   :stobjs (x86-32)))
   (declare (type (unsigned-byte 32) addr))
   (let* ((byte-num     (n02 addr))
          (word-addr    (ash addr -2))
          (dword        (memi word-addr x86-32))
          (shift-amount (ash byte-num 3))
          ;; Next form causes (callq (@ .SPBUILTIN-ASH)).
          (shifted-word (ash dword (- shift-amount))))
     (declare (type (unsigned-byte  2) byte-num)
              (type (unsigned-byte 30) word-addr)
              (type (unsigned-byte 32) dword shifted-word)
              (type (integer 0     24) shift-amount))
     (n08 shifted-word)))

 (defun rm32 (addr x86-32)
   (declare (xargs :guard (n32p addr)
                   :stobjs (x86-32)))
   (declare (type (unsigned-byte 32) addr))
   (let* ((byte-num  (n02 addr))
          (word-addr (ash addr -2))
          (dword     (memi word-addr x86-32)))
     (declare (type (unsigned-byte  2) byte-num)
              (type (unsigned-byte 30) word-addr)
              (type (unsigned-byte 32) dword))
     (if (= byte-num 0)
         dword
       ;; Memory "wrap" possible
       (let* ((byte0 (rm08       addr    x86-32))
              (byte1 (rm08 (n32+ addr 1) x86-32))
              (byte2 (rm08 (n32+ addr 2) x86-32))
              (byte3 (rm08 (n32+ addr 3) x86-32)))
         (declare (type (unsigned-byte 8) byte0 byte1 byte2 byte3))
         (logior (ash byte3 24) (ash byte2 16) (ash byte1 8) byte0))))))

(defthm natp-rm08
  (implies (and (force (x86-32p x86-32))
                (force (n32p addr)))
           (and (integerp (rm08 addr x86-32))
                (<= 0 (rm08 addr x86-32))))
  :rule-classes :type-prescription)

(defthm rm08-less-than-expt-2-32
  (implies (and (x86-32p x86-32)
                (n32p addr))
           (< (rm08 addr x86-32) 256))
  :rule-classes :linear)

(in-theory (disable rm08))

(in-theory (disable rm08))

(encapsulate
 ()
 (local
  (def-gl-thm logior-ash-bytes-<=-0
    :hyp (and (n08p byte0)
              (n08p byte1)
              (n08p byte2)
              (n08p byte3))
    :concl
    (<= 0 (logior (ash byte3 24) (ash byte2 16) (ash byte1 8) byte0))
    :g-bindings
    `((byte0 (:g-number ,(gl-int  0 1 9)))
      (byte1 (:g-number ,(gl-int  9 1 9)))
      (byte2 (:g-number ,(gl-int 18 1 9)))
      (byte3 (:g-number ,(gl-int 27 1 9))))))

 (local
  (def-gl-thm logior-ash-bytes-<-4294967296
    :hyp (and (n08p byte0)
              (n08p byte1)
              (n08p byte2)
              (n08p byte3))
    :concl
    (< (logior (ash byte3 24) (ash byte2 16) (ash byte1 8) byte0)
       4294967296)
    :g-bindings
    `((byte0 (:g-number ,(gl-int  0 1 9)))
      (byte1 (:g-number ,(gl-int  9 1 9)))
      (byte2 (:g-number ,(gl-int 18 1 9)))
      (byte3 (:g-number ,(gl-int 27 1 9))))))

 (defthm integerp-rm32
   (implies (and (force (x86-32p x86-32))
                 (force (n32p addr)))
            (and (integerp (rm32 addr x86-32))
                 (<= 0 (rm32 addr x86-32))))
   :rule-classes :type-prescription)

 (defthm rm32-less-than-expt-2-32
   (implies (and (x86-32p x86-32)
                 (n32p addr))
            (< (rm32 addr x86-32) 4294967296))
   :rule-classes :linear))

(in-theory (disable rm32))

; Write memory byte, memory double-word functions

(encapsulate
 ()

 (local
  (def-gl-thm ash-addr--2-is-zero-or-more
    :hyp (n32p addr)
    :concl (<= 0 (ash addr -2))
    :g-bindings
    `((addr (:g-number ,(gl-int  0 1 33))))))

 (local
  (def-gl-thm ash-addr--2-is-less
    :hyp (n32p addr)
    :concl (< (ash addr -2) 1073741824)
    :g-bindings
    `((addr (:g-number ,(gl-int  0 1 33))))))

 (local
  (def-gl-thm ash-logand-addr-3-is-integerp-less-than-or-equal-24
    :hyp (n32p addr)
    :concl (<= (ash (logand addr 3) 3) 24)
    :g-bindings
    `((addr (:g-number ,(gl-int  0 1 33))))))

 (local
  (def-gl-thm n32p-ash-x-is-less-than-n30p
    :hyp (n32p addr)
    :concl (< (ash addr (- 2)) 1073741824)
    :rule-classes :linear
    :g-bindings
    `((addr (:g-number ,(gl-int  0 1 33))))))

 (local
  (def-gl-thm ash-n08p-ash-logand-addr-3-3
    :hyp (and (n32p addr)
              (n08p byte))
    :concl (< (ash byte (ash (logand addr 3) 3))
               4294967296)
    :g-bindings
    `((addr (:g-number ,(gl-int  0 1 33)))
      (byte (:g-number ,(gl-int 33 1 9))))))

 (local
  (def-gl-thm word-to-write-equality
    :hyp (and (n08p byte)
              (n32p val)
              (n32p addr))
    :concl
    (equal (logand (logior (logand (lognot (ash 255 (ash (logand addr 3) 3))) val)
                           (ash byte (ash (logand addr 3) 3)))
                   4294967295)
           (logior (logand (lognot (ash 255 (ash (logand addr 3) 3)))
                           val)
                   (ash byte (ash (logand addr 3) 3))))
    :g-bindings
    `((addr (:g-number ,(gl-int  0 1 33)))
      (byte (:g-number ,(gl-int 33 1 9)))
      (val  (:g-number ,(gl-int 42 1 33))))))

 (local
  (def-gl-thm natp-word-to-write
    :hyp (and (n08p byte)
              (n32p value)
              (n32p addr))
    :concl
    (<= 0 (logior (logand (lognot (ash 255 (ash (logand addr 3) 3))) value)
                  (ash byte (ash (logand addr 3) 3))))
    :rule-classes :linear
    :g-bindings
    `((addr  (:g-number ,(gl-int  0 1 33)))
      (byte  (:g-number ,(gl-int 33 1 9)))
      (value (:g-number ,(gl-int 42 1 33))))))

 (local
  (def-gl-thm word-to-write-equality-less-than-2^32
    :hyp (and (n08p byte)
              (n32p val)
              (n32p addr))
    :concl
    (< (logior (logand (lognot (ash 255 (ash (logand addr 3) 3))) val)
               (ash byte (ash (logand addr 3) 3)))
       4294967296)
    :rule-classes :linear
    :g-bindings
    `((addr (:g-number ,(gl-int  0 1 33)))
      (byte (:g-number ,(gl-int 33 1 9)))
      (val  (:g-number ,(gl-int 42 1 33))))))

 (local
  (def-gl-thm another-logand-fact
    :hyp (and (n32p word)
              (n08p byte)
              (n32p addr))
    :concl
    (<= 0 (logior (logand (+ -1 (- (ash 255 (ash (logand addr 3) 3)))) word)
                  (ash byte (ash (logand addr 3) 3))))
    :g-bindings
    `((addr (:g-number ,(gl-int  0 1 33)))
      (byte (:g-number ,(gl-int 33 1 9)))
      (word (:g-number ,(gl-int 42 1 33))))))

 (defun wm08 (addr byte x86-32)
   (declare (xargs :guard (and (n32p addr)
                               (n08p byte))
                   :stobjs (x86-32)
                   :guard-hints
                   (("Goal" :in-theory (e/d () (lognot))))))
   (declare (type (integer 0 4294967296) addr))
   (let* ((byte-num      (n02 addr))
          (word-addr     (ash addr -2))
          (dword         (memi word-addr x86-32))
          (mask          #xFF)
          (shift-amount  (ash byte-num 3))
          (byte-mask     (ash mask shift-amount))
          (dword-masked  (logand (lognot byte-mask) dword))
          (byte-to-write (ash byte shift-amount))
          (word-to-write (logior dword-masked byte-to-write)))
     (declare (type (unsigned-byte  2) byte-num)
              (type (integer 0     24) shift-amount)
              (type (unsigned-byte 30) word-addr)
              (type (unsigned-byte 32)
                    byte-mask dword dword-masked byte-to-write))
     (update-memi word-addr word-to-write x86-32)))

 (defthm x86-32p-wm08
   (implies (and (x86-32p x86-32)
                 (n32p addr)
                 (n08p byte))
            (x86-32p (wm08 addr byte x86-32)))))

(in-theory (disable wm08))

(encapsulate
 ()
 (local
  (def-gl-thm ash-addr--2-is-zero-or-more
    :hyp (n32p addr)
    :concl (<= 0 (ash addr -2))
    :g-bindings
    `((addr (:g-number ,(gl-int  0 1 33))))))

 (local
  (def-gl-thm ash-addr--2-is-less
    :hyp (n32p addr)
    :concl (< (ash addr -2) 1073741824)
    :rule-classes :linear
    :g-bindings
    `((addr (:g-number ,(gl-int  0 1 33))))))

 (defun wm32 (addr dword x86-32)
   (declare (xargs :guard (and (n32p addr)
                               (n32p dword))
                   :stobjs (x86-32)))
   (declare (type (integer 0 4294967296) addr))
   (let* ((byte-num  (n02 addr))
          (word-addr (ash addr -2)))
     (declare (type (integer 0          4) byte-num)
              (type (integer 0 1073741824) word-addr))
     (if (= byte-num 0)
         (update-memi word-addr dword x86-32)
       ;; Memory "wrap" possible
       (b* ((x86-32 (wm08       addr    (n08      dword     ) x86-32))
            (x86-32 (wm08 (n32+ addr 1) (n08 (ash dword  -8)) x86-32))
            (x86-32 (wm08 (n32+ addr 2) (n08 (ash dword -16)) x86-32))
            (x86-32 (wm08 (n32+ addr 3) (n08 (ash dword -24)) x86-32)))
           x86-32))))

 (defthm x86-32p-wm32
   (implies (and (x86-32p x86-32)
                 (n32p addr)
                 (n32p dword))
            (x86-32p (wm32 addr dword x86-32)))))

(in-theory (disable wm32))


